@@ -14,13 +14,16 @@ except ImportError:  # python 2.6
     from . import argparse_compat as argparse
 import os
 import pwd
+import re
 import ssl
 import sys
 import textwrap
+import shlex
 
 from gunicorn import __version__
 from gunicorn import _compat
 from gunicorn.errors import ConfigError
+from gunicorn.reloader import reloader_engines
 from gunicorn import six
 from gunicorn import util
 
@@ -28,12 +31,6 @@ KNOWN_SETTINGS = []
 PLATFORM = sys.platform
 
 
-def wrap_method(func):
-    def _wrapped(instance, *args, **kwargs):
-        return func(*args, **kwargs)
-    return _wrapped
-
-
 def make_settings(ignore=None):
     settings = {}
     ignore = ignore or ()
@@ -45,6 +42,13 @@ def make_settings(ignore=None):
     return settings
 
 
+def auto_int(_, x):
+    # for compatible with octal numbers in python3
+    if re.match(r'0(\d)', x, re.IGNORECASE):
+        x = x.replace('0', '0o', 1)
+    return int(x, 0)
+
+
 class Config(object):
 
     def __init__(self, usage=None, prog=None):
@@ -68,6 +72,11 @@ class Config(object):
             raise AttributeError("No configuration setting for: %s" % name)
         self.settings[name].set(value)
 
+    def get_cmd_args_from_env(self):
+        if 'GUNICORN_CMD_ARGS' in self.env_orig:
+            return shlex.split(self.env_orig['GUNICORN_CMD_ARGS'])
+        return []
+
     def parser(self):
         kwargs = {
             "usage": self.usage,
@@ -110,14 +119,6 @@ class Config(object):
             worker_class.setup()
         return worker_class
 
-    @property
-    def threads(self):
-        return self.settings['threads'].get()
-
-    @property
-    def workers(self):
-        return self.settings['workers'].get()
-
     @property
     def address(self):
         s = self.settings['bind'].get()
@@ -203,6 +204,29 @@ class Config(object):
 
         return True
 
+    @property
+    def reuse_port(self):
+        return self.settings['reuse_port'].get()
+
+    @property
+    def paste_global_conf(self):
+        raw_global_conf = self.settings['raw_paste_global_conf'].get()
+        if raw_global_conf is None:
+            return None
+
+        global_conf = {}
+        for e in raw_global_conf:
+            s = _compat.bytes_to_str(e)
+            try:
+                k, v = re.split(r'(?<!\\)=', s, 1)
+            except ValueError:
+                raise RuntimeError("environment setting %r invalid" % s)
+            k = k.replace('\\=', '=')
+            v = v.replace('\\=', '=')
+            global_conf[k] = v
+
+        return global_conf
+
 
 class SettingMeta(type):
     def __new__(cls, name, bases, attrs):
@@ -212,7 +236,7 @@ class SettingMeta(type):
             return super_new(cls, name, bases, attrs)
 
         attrs["order"] = len(KNOWN_SETTINGS)
-        attrs["validator"] = wrap_method(attrs["validator"])
+        attrs["validator"] = staticmethod(attrs["validator"])
 
         new_class = super_new(cls, name, bases, attrs)
         new_class.fmt_desc(attrs.get("desc", ""))
@@ -334,6 +358,14 @@ def validate_string(val):
     return val.strip()
 
 
+def validate_file_exists(val):
+    if val is None:
+        return None
+    if not os.path.exists(val):
+        raise ValueError("File %s does not exists." % val)
+    return val
+
+
 def validate_list_string(val):
     if not val:
         return []
@@ -345,6 +377,10 @@ def validate_list_string(val):
     return [validate_string(v) for v in val]
 
 
+def validate_list_of_existing_files(val):
+    return [validate_file_exists(v) for v in validate_list_string(val)]
+
+
 def validate_string_to_list(val):
     val = validate_string(val)
 
@@ -380,7 +416,7 @@ def validate_callable(arity):
                     "" % (obj_name, mod_name))
         if not six.callable(val):
             raise TypeError("Value is not six.callable: %s" % val)
-        if arity != -1 and arity != len(inspect.getargspec(val)[0]):
+        if arity != -1 and arity != _compat.get_arity(val):
             raise TypeError("Value must have an arity of: %s" % arity)
         return val
     return _validate_callable
@@ -418,7 +454,7 @@ def validate_group(val):
 def validate_post_request(val):
     val = validate_callable(-1)(val)
 
-    largs = len(inspect.getargspec(val)[0])
+    largs = _compat.get_arity(val)
     if largs == 4:
         return val
     elif largs == 3:
@@ -443,22 +479,6 @@ def validate_chdir(val):
     return path
 
 
-def validate_file(val):
-    if val is None:
-        return None
-
-    # valid if the value is a string
-    val = validate_string(val)
-
-    # transform relative paths
-    path = os.path.abspath(os.path.normpath(os.path.join(util.getcwd(), val)))
-
-    # test if the path exists
-    if not os.path.exists(path):
-        raise ConfigError("%r not found" % val)
-
-    return path
-
 def validate_hostport(val):
     val = validate_string(val)
     if val is None:
@@ -469,6 +489,14 @@ def validate_hostport(val):
     else:
         raise TypeError("Value must consist of: hostname:port")
 
+
+def validate_reload_engine(val):
+    if val not in reloader_engines:
+        raise ConfigError("Invalid reload_engine: %r" % val)
+
+    return val
+
+
 def get_default_config_file():
     config_path = os.path.join(os.path.abspath(os.getcwd()),
             'gunicorn.conf.py')
@@ -577,22 +605,31 @@ class WorkerClass(Setting):
 
         The default class (``sync``) should handle most "normal" types of
         workloads. You'll want to read :doc:`design` for information on when
-        you might want to choose one of the other worker classes.
+        you might want to choose one of the other worker classes. Required
+        libraries may be installed using setuptools' ``extra_require`` feature.
 
         A string referring to one of the following bundled classes:
 
         * ``sync``
-        * ``eventlet`` - Requires eventlet >= 0.9.7
-        * ``gevent``   - Requires gevent >= 0.13
-        * ``tornado``  - Requires tornado >= 0.2
+        * ``eventlet`` - Requires eventlet >= 0.9.7 (or install it via 
+          ``pip install gunicorn[eventlet]``)
+        * ``gevent``   - Requires gevent >= 0.13 (or install it via 
+          ``pip install gunicorn[gevent]``)
+        * ``tornado``  - Requires tornado >= 0.2 (or install it via 
+          ``pip install gunicorn[tornado]``)
         * ``gthread``  - Python 2 requires the futures package to be installed
-        * ``gaiohttp`` - Requires Python 3.4 and aiohttp >= 0.21.5
+          (or install it via ``pip install gunicorn[gthread]``)
+        * ``gaiohttp`` - Deprecated.
 
         Optionally, you can provide your own worker by giving Gunicorn a
         Python path to a subclass of ``gunicorn.workers.base.Worker``.
         This alternative syntax will load the gevent class:
-        ``gunicorn.workers.ggevent.GeventWorker``. Alternatively, the syntax
-        can also load the gevent class with ``egg:gunicorn#gevent``.
+        ``gunicorn.workers.ggevent.GeventWorker``.
+
+        .. deprecated:: 19.8
+           The ``gaiohttp`` worker is deprecated. Please use
+           ``aiohttp.worker.GunicornWebWorker`` instead. See
+           :ref:`asyncio-workers` for more information on how to use it.
         """
 
 class WorkerThreads(Setting):
@@ -612,7 +649,14 @@ class WorkerThreads(Setting):
         You'll want to vary this a bit to find the best for your particular
         application's work load.
 
-        If it is not defined, the default is 1.
+        If it is not defined, the default is ``1``.
+
+        This setting only affects the Gthread worker type.
+        
+        .. note::
+           If you try to use the ``sync`` worker type and set the ``threads``
+           setting to more than 1, the ``gthread`` worker type will be used
+           instead.
         """
 
 
@@ -716,7 +760,14 @@ class Keepalive(Setting):
     desc = """\
         The number of seconds to wait for requests on a Keep-Alive connection.
 
-        Generally set in the 1-5 seconds range.
+        Generally set in the 1-5 seconds range for servers with direct connection
+        to the client (e.g. when you don't have separate load balancer). When
+        Gunicorn is deployed behind a load balancer, it often makes sense to
+        set this to a higher value.
+
+        .. note::
+           ``sync`` worker does not support persistent connections and will
+           ignore this option.
         """
 
 
@@ -773,8 +824,12 @@ class LimitRequestFieldSize(Setting):
     desc = """\
         Limit the allowed size of an HTTP request header field.
 
-        Value is a number from 0 (unlimited) to 8190. to set the limit
-        on the allowed size of an HTTP request header field.
+        Value is a positive number or 0. Setting it to 0 will allow unlimited
+        header field sizes.
+
+        .. warning::
+           Setting this parameter to a very high or unlimited value can open
+           up for DDOS attacks.
         """
 
 
@@ -785,6 +840,7 @@ class Reload(Setting):
     validator = validate_bool
     action = 'store_true'
     default = False
+
     desc = '''\
         Restart workers when code changes.
 
@@ -794,9 +850,53 @@ class Reload(Setting):
         The reloader is incompatible with application preloading. When using a
         paste configuration be sure that the server block does not import any
         application code or the reload will not work as designed.
+
+        The default behavior is to attempt inotify with a fallback to file
+        system polling. Generally, inotify should be preferred if available
+        because it consumes less system resources.
+
+        .. note::
+           In order to use the inotify reloader, you must have the ``inotify``
+           package installed.
         '''
 
 
+class ReloadEngine(Setting):
+    name = "reload_engine"
+    section = "Debugging"
+    cli = ["--reload-engine"]
+    meta = "STRING"
+    validator = validate_reload_engine
+    default = "auto"
+    desc = """\
+        The implementation that should be used to power :ref:`reload`.
+
+        Valid engines are:
+
+        * 'auto'
+        * 'poll'
+        * 'inotify' (requires inotify)
+
+        .. versionadded:: 19.7
+        """
+
+
+class ReloadExtraFiles(Setting):
+    name = "reload_extra_files"
+    action = "append"
+    section = "Debugging"
+    cli = ["--reload-extra-file"]
+    meta = "FILES"
+    validator = validate_list_of_existing_files
+    default = []
+    desc = """\
+        Extends :ref:`reload` option to also watch and reload on additional files
+        (e.g., templates, configurations, specifications, etc.).
+
+        .. versionadded:: 19.8
+        """
+
+
 class Spew(Setting):
     name = "spew"
     section = "Debugging"
@@ -814,7 +914,7 @@ class Spew(Setting):
 class ConfigCheck(Setting):
     name = "check_config"
     section = "Debugging"
-    cli = ["--check-config", ]
+    cli = ["--check-config"]
     validator = validate_bool
     action = "store_true"
     default = False
@@ -863,6 +963,21 @@ class Sendfile(Setting):
         """
 
 
+class ReusePort(Setting):
+    name = "reuse_port"
+    section = "Server Mechanics"
+    cli = ["--reuse-port"]
+    validator = validate_bool
+    action = "store_true"
+    default = False
+
+    desc = """\
+        Set the ``SO_REUSEPORT`` flag on the listening socket.
+
+        .. versionadded:: 19.8
+        """
+
+
 class Chdir(Setting):
     name = "chdir"
     section = "Server Mechanics"
@@ -932,6 +1047,14 @@ class WorkerTmpDir(Setting):
         A directory to use for the worker heartbeat temporary file.
 
         If not set, the default temporary directory will be used.
+
+        .. note::
+           The current heartbeat system involves calling ``os.fchmod`` on
+           temporary file handlers and may block a worker for arbitrary time
+           if the directory is on a disk-backed filesystem.
+
+           See :ref:`blocking-os-fchmod` for more detailed information
+           and a solution for avoiding this problem.
         """
 
 
@@ -966,14 +1089,13 @@ class Group(Setting):
         change the worker processes group.
         """
 
-
 class Umask(Setting):
     name = "umask"
     section = "Server Mechanics"
     cli = ["-m", "--umask"]
     meta = "INT"
     validator = validate_pos_int
-    type = int
+    type = auto_int
     default = 0
     desc = """\
         A bit mask for the file mode on files written by Gunicorn.
@@ -987,6 +1109,23 @@ class Umask(Setting):
         """
 
 
+class Initgroups(Setting):
+    name = "initgroups"
+    section = "Server Mechanics"
+    cli = ["--initgroups"]
+    validator = validate_bool
+    action = 'store_true'
+    default = False
+
+    desc = """\
+        If true, set the worker process's group access list with all of the
+        groups of which the specified username is a member, plus the specified
+        group id.
+
+        .. versionadded:: 19.7
+        """
+
+
 class TmpUploadDir(Setting):
     name = "tmp_upload_dir"
     section = "Server Mechanics"
@@ -1060,9 +1199,22 @@ class AccessLog(Setting):
     desc = """\
         The Access log file to write to.
 
-        ``'-'`` means log to stderr.
+        ``'-'`` means log to stdout.
         """
 
+class DisableRedirectAccessToSyslog(Setting):
+    name = "disable_redirect_access_to_syslog"
+    section = "Logging"
+    cli = ["--disable-redirect-access-to-syslog"]
+    validator = validate_bool
+    action = 'store_true'
+    default = False
+    desc = """\
+    Disable redirect access logs to syslog.
+
+    .. versionadded:: 19.8
+    """
+
 
 class AccessLogFormat(Setting):
     name = "access_log_format"
@@ -1074,30 +1226,31 @@ class AccessLogFormat(Setting):
     desc = """\
         The access log format.
 
-        ==========  ===========
-        Identifier  Description
-        ==========  ===========
-        h           remote address
-        l           ``'-'``
-        u           user name
-        t           date of the request
-        r           status line (e.g. ``GET / HTTP/1.1``)
-        m           request method
-        U           URL path without query string
-        q           query string
-        H           protocol
-        s           status
-        B           response length
-        b           response length or ``'-'`` (CLF format)
-        f           referer
-        a           user agent
-        T           request time in seconds
-        D           request time in microseconds
-        L           request time in decimal seconds
-        p           process ID
-        {Header}i   request header
-        {Header}o   response header
-        ==========  ===========
+        ===========  ===========
+        Identifier   Description
+        ===========  ===========
+        h            remote address
+        l            ``'-'``
+        u            user name
+        t            date of the request
+        r            status line (e.g. ``GET / HTTP/1.1``)
+        m            request method
+        U            URL path without query string
+        q            query string
+        H            protocol
+        s            status
+        B            response length
+        b            response length or ``'-'`` (CLF format)
+        f            referer
+        a            user agent
+        T            request time in seconds
+        D            request time in microseconds
+        L            request time in decimal seconds
+        p            process ID
+        {Header}i    request header
+        {Header}o    response header
+        {Variable}e  environment variable
+        ===========  ===========
         """
 
 
@@ -1147,7 +1300,7 @@ class CaptureOutput(Setting):
     action = 'store_true'
     default = False
     desc = """\
-        Redirect stdout/stderr to Error log.
+        Redirect stdout/stderr to specified file in :ref:`errorlog`.
 
         .. versionadded:: 19.6
         """
@@ -1166,10 +1319,8 @@ class LoggerClass(Setting):
         The default class (``gunicorn.glogging.Logger``) handle most of
         normal usages in logging. It provides error and access logging.
 
-        You can provide your own worker by giving Gunicorn a
+        You can provide your own logger by giving Gunicorn a
         Python path to a subclass like ``gunicorn.glogging.Logger``.
-        Alternatively the syntax can also load the Logger class
-        with ``egg:gunicorn#simple``.
         """
 
 
@@ -1187,6 +1338,24 @@ class LogConfig(Setting):
     """
 
 
+class LogConfigDict(Setting):
+    name = "logconfig_dict"
+    section = "Logging"
+    cli = ["--log-config-dict"]
+    validator = validate_dict
+    default = {}
+    desc = """\
+    The log config dictionary to use, using the standard Python
+    logging module's dictionary configuration format. This option
+    takes precedence over the :ref:`logconfig` option, which uses the
+    older file configuration format.
+
+    Format: https://docs.python.org/3/library/logging.config.html#logging.config.dictConfig
+
+    .. versionadded:: 19.8
+    """
+
+
 class SyslogTo(Setting):
     name = "syslog_addr"
     section = "Logging"
@@ -1226,6 +1395,11 @@ class Syslog(Setting):
     default = False
     desc = """\
     Send *Gunicorn* logs to syslog.
+
+    .. versionchanged:: 19.8
+
+     You can now disable sending access logs by using the
+     :ref:`disable-redirect-access-to-syslog` setting.
     """
 
 
@@ -1331,23 +1505,6 @@ class DefaultProcName(Setting):
         """
 
 
-class DjangoSettings(Setting):
-    name = "django_settings"
-    section = "Django"
-    cli = ["--settings"]
-    meta = "STRING"
-    validator = validate_string
-    default = None
-    desc = """\
-        The Python path to a Django settings module. (deprecated)
-
-        e.g. ``myproject.settings.main``. If this isn't provided, the
-        ``DJANGO_SETTINGS_MODULE`` environment variable will be used.
-
-        **DEPRECATED**: use the ``--env`` argument instead.
-        """
-
-
 class PythonPath(Setting):
     name = "pythonpath"
     section = "Server Mechanics"
@@ -1567,6 +1724,25 @@ class PostRequest(Setting):
         """
 
 
+class ChildExit(Setting):
+    name = "child_exit"
+    section = "Server Hooks"
+    validator = validate_callable(2)
+    type = six.callable
+
+    def child_exit(server, worker):
+        pass
+    default = staticmethod(child_exit)
+    desc = """\
+        Called just after a worker has been exited, in the master process.
+
+        The callable needs to accept two instance variables for the Arbiter and
+        the just-exited Worker.
+
+        .. versionadded:: 19.7
+        """
+
+
 class WorkerExit(Setting):
     name = "worker_exit"
     section = "Server Hooks"
@@ -1577,7 +1753,7 @@ class WorkerExit(Setting):
         pass
     default = staticmethod(worker_exit)
     desc = """\
-        Called just after a worker has been exited.
+        Called just after a worker has been exited, in the worker process.
 
         The callable needs to accept two instance variables for the Arbiter and
         the just-exited Worker.
@@ -1688,9 +1864,13 @@ class SSLVersion(Setting):
     section = "SSL"
     cli = ["--ssl-version"]
     validator = validate_pos_int
-    default = ssl.PROTOCOL_TLSv1
+    default = ssl.PROTOCOL_SSLv23
     desc = """\
     SSL version to use (see stdlib ssl module's)
+
+    .. versionchanged:: 19.7
+       The default value has been changed from ``ssl.PROTOCOL_TLSv1`` to
+       ``ssl.PROTOCOL_SSLv23``.
     """
 
 class CertReqs(Setting):
@@ -1747,3 +1927,25 @@ if sys.version_info >= (2, 7):
         desc = """\
         Ciphers to use (see stdlib ssl module's)
         """
+
+
+class PasteGlobalConf(Setting):
+    name = "raw_paste_global_conf"
+    action = "append"
+    section = "Server Mechanics"
+    cli = ["--paste-global"]
+    meta = "CONF"
+    validator = validate_list_string
+    default = []
+
+    desc = """\
+        Set a PasteDeploy global config variable in ``key=value`` form.
+
+        The option can be specified multiple times.
+
+        The variables are passed to the the PasteDeploy entrypoint. Example::
+
+            $ gunicorn -b 127.0.0.1:8000 --paste development.ini --paste-global FOO=1 --paste-global BAR=2
+
+        .. versionadded:: 19.7
+        """
